
from esrapyPatterns import *

class Line(Pattern):
	"""An indented pattern.

	This is some amount of white space preceeding a given Pattern.

	The returned Match is annotated with the exact indentation
	string so it can be compared by higher levels to determine block
	boundaries.

	Note the encapsulated Pattern should handle blank line and comment
	consumption, and should end by consuming the end of line delimiter,
	so that a raw succession of Line() patterns makes sense.

	Don't forget to include \Z as an option in your end-of-line match.
	"""
	name = "Indent.Line"

	space = Regex(r"[ \t]*")

	def init(self, pattern):
		self.pattern = pattern

	def spawn(self, parser, pos):
		parser.spawn(Line.space, pos, self.notify1, pos)

	def notify1(self, startPos, parser, match):
		indent = match.parsings[0][0].parsings[0]
		parser.spawn(self.pattern, match.end, self.notify2, (startPos, indent))

	def notify2(self, data, parser, match):
		startPos, indent = data
		parser.notify(self, startPos, match.end, (match,), indent)

	def compact(self, parsing, **flags):
		return parsing[0].compact(**flags)

	def str(self):
		return ">>"+deepStr(self.pattern)

class Block(Pattern):
	"""A block of identically indented identical patterns (at least one).

	In addition to enforcing equal indenting, the returned Match
	is annotated with the block indent string so that parent blocks
	can verify proper nesting.

	The passed pattern should be appropriate for Line(), above.
	"""
	name = "Indent.Block"

	def init(self, pattern):
		self.pattern = Line(pattern)

	def spawn(self, parser, pos):
		parser.spawn(self.pattern, pos, self.notify, ())

	def notify(self, data, parser, match):
		# Ignore this match if it's a mismatch to an existing indention precident:
		if data and match.attrs != data[0].attrs:
			return
		parsing = data + (match,)
		parser.spawn(self.pattern, match.end, self.notify, parsing)
		parser.notify(self, parsing[0].start, match.end, parsing, match.attrs)

	def compact(self, parsing, **flags):
		return [m.compact(**flags) for m in parsing]

	def str(self):
		return "Indent.Block[" + deepStr(self.pattern) + "]"

class Indent(Pattern):
	"""A header line with an indented body.

	Like a Line, the returned Match is annotated with the block indent string.

	Both the header and the body pattern should be appropriate for Line(), above.
	"""
	name = "Indent.Indent"

	def init(self, headPattern, bodyPattern):
		self.head = Line(headPattern)
		self.body = Block(bodyPattern)

	def spawn(self, parser, pos):
		parser.spawn(self.head, pos, self.notifyHead, None)

	def notifyHead(self, data, parser, match):
		parser.spawn(self.body, match.end, self.notifyBody, match)

	def notifyBody(self, headMatch, parser, bodyMatch):
		# Verify that the body is indented from the head:
		if bodyMatch.attrs.startswith(headMatch.attrs):
			parser.notify(self, headMatch.start, bodyMatch.end, (headMatch, bodyMatch), headMatch.attrs)
		# Could possibly print an error here, but only if we know there's no competing interpretation.

	def str(self):
		return "Indent[" + deepStr(self.pattern) + "]"

